//
//  This file is part of SienaXML, an XML interface to the 
//  Siena wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Author: Chris Corliss <corliss@falcon.stortek.com>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2002 University of Colorado
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: XMLProcessor.java,v 1.1.1.1 2002/06/07 22:42:37 jgiacomo Exp $
//
/* Generated by Together */

package siena.xml;

import java.util.HashMap;
import java.util.Vector;
import java.io.FileInputStream;
import org.xml.sax.InputSource;
import org.w3c.dom.Node;
import org.w3c.dom.Attr;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;

import siena.xml.rules.Rule;
import siena.xml.rules.RulesGenerator;

import org.apache.xerces.dom.ElementImpl;
import org.apache.xerces.dom.CoreDocumentImpl;
import java.util.Set;
import junit.framework.*;

/**
 * This class handles the processing of the XML notification and turning it
 * into a Siena notification. Before it can do anything with the XML
 * notification it has to be given the XSchema for the XML notification. Once
 * this is done then an XML notification can be translated. If the user has any
 * specific ways that they want the XML notification to be handled they can
 * submit a filename that contains the rules they want to be applied.
 *
 * @author Chris Corliss
 * @version 0.1
 * @see Rules
 * @since 11-05-2001
 */

public class XMLProcessor {
    
    /**
     * This is the parser to be used to parse the XML document and turn it
     * into a W3C DOM.
     * @since 11-05-2001
     */
    
    private DOMParserWrapper parser;
    
    /**
     * This HashMap contains the mapping between an attribute in Siena and its
     * value. This is used to store the different parts of the notification as
     * the XML notification is having the rules being applied to it. Once the
     * rules have been applied this HashMap will contain the complete
     * notiification to be sent.
     * @since 11-05-2001
     */
    
    private HashMap attributeValue;
    
    /**
     * This variable is used to hold the rules that are to be applied to the XML
     * tag.
     */
    
    private HashMap xmlRules = null;
    
    /**
     * This map holds the quantity attribute name and its value for an XML tag.
     * Only XML tags that are not unique will have values in this table.
     */
    
    private HashMap quantValues = new HashMap();
    
    /**
     * This map holds the user rules. Unlike the xmlRules hash map these rules
     * still have to be applied to the node.
     */
    
    private HashMap userRules = null;
    
    /**
     * This contains the minimum and maximum numerical values for XML tags that
     * have numerical data.
     */
    
    private HashMap minMaxValues = new HashMap();
    
    /**
     * This is the Schema processor used to parse the XSchema and produce the
     * xmlRules.
     */
    
    private XSchemaProcessor processor;
    
    /**
     * This FileWriter is used to log warning messages when a possible problem
     * could occur using the user rules.
     */
    
    private java.io.FileWriter logFile;
    
    /**
     * This is the file name of the log file which will contain the warning
     * messages.
     */
    
    private String logfileName = "RuleErrorLog.txt";
    
    /**
     * Public constructor
     * @since 11-05-2001
     * @stereotype constructor
     */
    
    public XMLProcessor() {
        parser = new DOMParserWrapper();
        xmlRules = new HashMap();
        attributeValue = new HashMap();
        processor = new XSchemaProcessor();

	// Open the Log file for translation warnings.
        try {
            logFile = new java.io.FileWriter(logfileName, true);
        } catch (java.io.IOException exception) {
            System.err.println("Unable to open a log file: "+logfileName);
            System.err.println("Exception is: "+exception.getMessage());
            logFile = null;
        }
    }
    
    /**
     * This method will set the name of the log file which is used to write to
     * when multiple user rules can match a single XML tag.
     *
     * @param fileName Name of the file to be used as the log file.
     */
    
    public void setLogFileName(String fileName) {
        
        // If the log file exists, try to close it.
	 
        if (logFile != null) {
            try {
                logFile.flush(); // Make sure everything has been written first.
                logFile.close();
            } catch (java.io.IOException exception) {
                System.err.println("Unable to close previous logfile: "+logfileName);
                System.err.println("Exception: "+exception.getMessage());
            }
        }
        
        // Create a new log file.
        try {
            logFile = new java.io.FileWriter(fileName, true);
        } catch (java.io.IOException exception) {
            System.err.println("Unable to open a log file: "+logfileName);
            System.err.println("Exception is: "+exception.getMessage());
            logFile = null;
        }
        
        logfileName = fileName;
    }
    
    /**
     * This method will set the location of the style sheet.
     *
     * @param stylesheet Name of the file that contains the XSchema
     * @throws SienaException Thrown when the file can not be accessed or parsed correctly.
     */
    
    public void setStyleSheet(String stylesheet) throws siena.SienaException {
        try {
            processor.setXSchema(stylesheet);
        } catch (Exception exception) {
            throw new siena.SienaException("Unable to set XSchema:"+exception.getMessage());
        }
        xmlRules = processor.getXMLRules();
    }
    
    /**
     * This method translates an XML notification into a Siena notification.
     *
     * @return Siena Notification object to be sent to the Siena server
     * @since 11-05-2001
     * @param xmlNotification String containing the the XML Notification to be sent
     * @throws SienaException If the XML notification can not be translated this exception is thrown.
     */
    
    public siena.Notification translate(String xmlNotification) throws siena.SienaException {
          
        // This is the root node of the XML document
        Node document = null;
        siena.Notification notification = null;

        // Used to hold the zipped bytes of the XML notification
        java.io.ByteArrayOutputStream bytes = null;

        // This is used to write the XML notification into the byte array.
        java.io.ObjectOutputStream zipOutput = null;
        
        /* Parse the XML notification passed in. If there is a problem throw an
         * exception.
         */
        try {
            document = parser.parse(xmlNotification);
            document.normalize();
        } catch (Exception e) {
            throw new siena.SienaException(e.getMessage());
        }
        
        applyRules(document,"");
        
        notification = new siena.Notification();
        
        populateNotification(notification);
        
        generateMinMaxValues(notification);
        
        generateQuantityValues(notification);
        
        // Zip the XML Notification and store it in a byte array. This byte array will
        // be put into the siena notification.
        xmlNotification.getBytes();
        bytes = new java.io.ByteArrayOutputStream();
        try {
            zipOutput = new java.io.ObjectOutputStream(new java.util.zip.GZIPOutputStream(bytes));
            zipOutput.writeUTF(xmlNotification);
            zipOutput.close();
        } catch (java.io.IOException exception) {
            System.err.println("exception: "+exception);
            System.err.println("Notification not sent!");
            return null;
        }
        
        notification.putAttribute("_xml", bytes.toByteArray());
        
        return notification;
    }
    
    
    /**
     * This method will populate the notification with the names and values
     * stored in the attributeValue hash map.
     *
     * @param notification The notification to populate with the values.
     */
    
    private void populateNotification(siena.Notification notification) {
        Vector values = null;
        boolean isInteger = false;
        boolean isDouble = false;
        StringBuffer buffer = null;
        Object[] keys = null;
        Object value = null;
        java.util.Set keySet = null;
        
        keySet = attributeValue.keySet();
        keys = keySet.toArray();
        
        for (int i = 0; i < keys.length; i++) {
            // Reset the numeric flags each time the for loop is executed
            isInteger = false;
            isDouble = false;
            
            values = (Vector)attributeValue.get(keys[i]);
            buffer = new StringBuffer();
            
            for (int j = 0; j < values.size(); j++) {
                value = values.elementAt(j);
                if (value instanceof Integer) {
                    isInteger = true;
                    break;
                } else if (value instanceof Double) {
                    isDouble = true;
                    break;
                }
                
                /**
                 * Deliminate the multiple values for the attribute by a coma followed
                 * by a space. This should not be done on the first time through the
                 * loop since it could also be the last iteration through the loop.
                 */
                
                if (j != 0) {
                    buffer.append(", ");
                }
                buffer.append(values.elementAt(j).toString());
            }
            if (isInteger) {
                notification.putAttribute((String)keys[i], ((Integer)value).intValue());
            } else if (isDouble) {
                notification.putAttribute((String)keys[i], ((Double)value).doubleValue());
            } else {
                notification.putAttribute((String)keys[i], buffer.toString());
            }
        }
        
        // This clears the HashMap so that the data from this translation will
        // not carry over into the next translation.
        attributeValue = new HashMap();
    }
    
    /**
     * This method will generate the min max values for all of the XML tags that
     * contain numeric data in them.
     *
     * @param notification The notification to store the min max attributes in.
     */
    
    private void generateMinMaxValues(siena.Notification notification) {
        MinMax temp = null;
        java.util.Set keySet = null;
        Object[] keys = null;
        
        keySet = minMaxValues.keySet();
        keys = keySet.toArray();
        for (int i = 0; i < keys.length; i++) {
            temp = (MinMax)minMaxValues.get(keys[i]);
            if (temp.usesInts()) {
                notification.putAttribute((String)keys[i]+"$min", temp.getIntMin());
                notification.putAttribute((String)keys[i]+"$max", temp.getIntMax());
            } else {
                notification.putAttribute((String)keys[i]+"$min", temp.getDoubleMin());
                notification.putAttribute((String)keys[i]+"$max", temp.getDoubleMax());
            }
        }
        
        // Clear out the values in the hash map so that they will not carry
        // over into the next translation.
        minMaxValues = new HashMap();
    }
    
    /**
     * This method will store the generated quantity information in the Siena
     * notification.
     *
     * @param notification The Siena notification to store the attributes in.
     */
    
    private void generateQuantityValues(siena.Notification notification) {
        Vector temp = null;
        java.util.Set keySet = null;
        Object[] keys = null;
        
        keySet = quantValues.keySet();
        keys = keySet.toArray();
        for (int i = 0; i < keys.length; i++) {
            temp = (Vector)quantValues.get(keys[i]);
            notification.putAttribute((String)keys[i], ((Integer)temp.elementAt(0)).intValue());
        }
        
        // Clear out the values in the hash map so that they will not carry
        // over into the next translation.
        quantValues = new HashMap();
    }
    
    /**
     * This private method is invoked from within the translate method. It
     * will apply the rules in processingRules or in modifiedRules, if
     * modifedRules is not null, to the XML notification. This is done by
     * tranversing the DOM tree, the root of which is passed into this method.
     * With each node the method checks to see if there is a rule associated
     * with the tag. If so it executes the rule otherwise the default behavior
     * is to transform the tag into a Siena attribute name with the value of the
     * tag becoming the value of the attribute.
     *
     * @since 11-05-2001
     * @param node The current node to process in the XML DOM tree.
     * @param prefix This is the prefix to be applied to the name of the attribute in the notification
     */
    
    private void applyRules(Node node, String prefix) {
        String currentPrefix = prefix;
        String newPrefix = prefix;
        String xmlTag = node.getLocalName();
        String xmlValue = null;
        String attribute = null;
        String userAttribute = null;
        String tempString = null;
        boolean ignorePrefix = false;
        boolean ignoreThisNode = false;
        int index = -1;
        Vector tempVector = null;
        String prefixAttr = null;
        Node attributeNode = null;
        Node tempNode = null;
        MinMax tempMM = null;
        Integer xmlIntegerValue = null;
        Double xmlDoubleValue = null;
        org.w3c.dom.NamedNodeMap attributes = null;
        NodeList children = null;
        
        // Because of the way the value for a node is gotten from it, the method
        // does not need to process the TEXT_NODEs. See DOMTreeProcessing.
        // getNodeValue.
        if (node.getNodeType() == Node.TEXT_NODE) {
            return;
        }
        
        xmlValue = DOMTreeProcessing.getNodeValue(node);
        
        if (xmlValue != null) {
            xmlValue = xmlValue.trim();
            if (xmlValue.length() == 0) {
                xmlValue = null;
            }
        }
        
        if (userRules != null) {
            userAttribute = tryToApplyUserRules(node);
            
            // If a user rule was applied to the node then in most cases the
            // user will not want the prefix to be enabled. However, if the rule
            // had a unique action, then simply clear out the prefix as it will
            // be used later.
            
            if (userAttribute != null) {
                if(userAttribute.indexOf('$') != -1) {
                    prefix="";
                } else {
                    ignorePrefix = true;
                }
            }
        }
        
        attribute = (String)xmlRules.get(xmlTag);
        
        // This is a check to see if a user rule does something different than a
        // default rule. If so a warning message is printed to the log file or ]
        // System.err if no log file exists.
        
        if (userAttribute != null && attribute != null) {
            if (!userAttribute.equals(attribute)) {
                if (logFile != null) {
                    try {
                        logFile.write("WARNING:\n");
                        logFile.write("A user rule produces a different output than a default rule.\n");
                        logFile.write("XML Tag: "+xmlTag+"\nDefault: "+attribute+"\nUser: "+userAttribute+"\n");
                        logFile.write("Note: User rule will be used instead of the default rule.\n\n");
                        logFile.flush();
                    } catch (java.io.IOException exception) {
                        System.err.println("Unable to write to the log file: "+logfileName);
                        System.err.println("Exception: "+exception.getMessage());
                    }
                } else {
                    System.err.println("WARNING: A user rule produces a different output than a default rule.");
                    System.err.println("XML Tag: "+xmlTag+"\nDefault: "+attribute+"\nUser: "+userAttribute+"\n");
                }
            }
        }

        // This is done to keep the rest of this function the same.
        if (userAttribute != null) {
            attribute = userAttribute;
        }
        
        // Deal with any processing insturctions in the value returned by the
        // map or user rule.
        if (attribute != null) {
            index = attribute.indexOf("$");
            
            // Deal with the 'unique' processing insturction
            if (index !=-1) {
                
                // the prefixAttr is the name of the attribute or child tag
                // which should be used to uniquely identify the current tag
                // that is being processed.
                prefixAttr = attribute.substring(index+1);
                attributeNode = node.getAttributes().getNamedItem(prefixAttr);
                
                // The unique item that was specified is not an attribute. It
                // is then a tag which needs to be a child of the current tag.
                if (attributeNode == null) {
                    attributeNode = DOMTreeProcessing.findChildByName(prefixAttr, node);
                }
                
                // If we can not find it continue processing. Otherwise determine
                // the unique name for the SIENA notification.
                if (attributeNode != null) {
                    attribute = attribute.substring(0, index-1);
                    
                    tempString = attributeNode.getNodeValue();
                    if (tempString == null) {
                        tempString = attributeNode.getFirstChild().getNodeValue();
                    }
                    
                    // Generate the new prefix to use
                    if (tempString != null) {
                        newPrefix += attribute+"/"+tempString.replace('-','.')+"/";
                    } else {
                        newPrefix += attribute+"/"+attributeNode.getNodeValue().replace('-','.')+"/";
                    }
                    
                    // Generate the attribute that holds the number of tags that
                    // have the same name as the current XML tag.
                    generateQuantityValue(attribute, node);
                }
                
                // Handle the IGNORE instruction
            } else if (attribute.equals("IGNORE")) {
                xmlValue = null;
                ignoreThisNode = true;
                
                // Handle the PATH insturction
            } else if (attribute.equals("PATH")){
                newPrefix += node.getLocalName()+"/";
                
                // No processing instructions, default case
            } else {
                xmlTag = attribute;
            }
        }
        
        /* This next section of code handles the processing of the value of the
         * xml tag. The first check is to see if the value is not null and if
         * we were not told to ignore this tag.
         */
        
        if (xmlValue != null && !ignoreThisNode) {
            // Check to see if the value for the node is numeric, either an
            // integer or double. IF so, update the MinMax object that is
            // keeping track of the minimum and maximum values for all of the
            // tags with the same name.
            
            // Check for Integer data....
            try {
                xmlIntegerValue = new Integer(xmlValue);
                tempMM = (MinMax)minMaxValues.get(xmlTag);
                if (tempMM == null) {
                    tempMM = new MinMax();
                }
                tempMM.adjust(xmlIntegerValue.intValue());
                minMaxValues.put(xmlTag, tempMM);
            } catch (Exception ignore) {
                // Nothing is done here since the code in the try block is
                // basically a test to see if the value is an integer.
            };
            
            // Check for double.....
            try {
                xmlDoubleValue = new Double(xmlValue);
                tempMM = (MinMax)minMaxValues.get(xmlTag);
                if (tempMM == null) {
                    tempMM = new MinMax();
                }
                tempMM.adjust(xmlDoubleValue.doubleValue());
                minMaxValues.put(xmlTag, tempMM);
                
            } catch (Exception ignore) {
                // Nothing is done here since the code is only a test and
                // nothing else.
            };
            
            // If this is the first time we are adding a value to this attribute
            // then we will need to create a new Vector to hold all of the
            // values for this attribute.
            tempVector = (Vector)attributeValue.get(xmlTag);
            if (tempVector == null) {
                tempVector = new Vector();
            } else {
                currentPrefix = "";
            }
            
            // Add the respective value to the Vector: if we had an integer or
            // a double then add the created Integer or Double objects. Otherwise
            // add the value in the xmlValue variable.
            if (xmlIntegerValue != null) {
                tempVector.addElement(xmlIntegerValue);
            } else if (xmlDoubleValue != null) {
                tempVector.addElement(xmlDoubleValue);
            } else {
                tempVector.addElement(xmlValue);
            }
            attributeValue.put(currentPrefix+xmlTag,tempVector);
        }
        
        // If we have not been asked to ignore this node then process the nodes
        // attributes then children.
        
        if (!ignoreThisNode) {
            if (node.hasAttributes()) {
                attributes = node.getAttributes();
                for (int i = 0; i < attributes.getLength(); i++) {
                    // If we used a user rule on this tag check to see if
                    // we are supposed to ignore the prefix. If so, send an
                    // empty string.
                    if (ignorePrefix) {
                        applyRules(attributes.item(i),"");
                    } else {
                        applyRules(attributes.item(i), currentPrefix);
                    }
                }
            }
            
            if (!(node instanceof org.w3c.dom.Attr)) {
                children = node.getChildNodes();
                for (int i=0; i < children.getLength(); i++) {
                    // If we used a user rule on this tag check to see if
                    // we are supposed to ignore the prefix. If so, send an
                    // empty string.
                    if (ignorePrefix) {
                        applyRules(children.item(i), "");
                    } else {
                        applyRules(children.item(i), newPrefix);
                    }
                }
            }
        }
    }
    
    /**
     * This method keeps a record of t he number of times a particular XML tag
     * appears in the XML notification when the XML tag is not unique. It will
     * update a HashMap that maps an XML tag to an Integer object that contains
     * the number of times the XML tag has been seen.
     *
     * @param name Name of the base notification attribute.
     * @param node The node to use in the check.
     */
    
    private void generateQuantityValue(String name, Node node) {
        String quantAttribute = null;
        int numberOfTags = 0;
        Vector vector = null;
        String attributeQuantityName = name+"$quant";
        String key = null;
        java.util.Set keySet = null;
        Object[] keys = null;
        
        keySet = quantValues.keySet();
        keys = keySet.toArray();
        
        // Try to see if we have already started to count the number of times
        // the XML tag has occured. This is done by looking through the keys of
        // the hash map used to keep record of XML tags and the number of times
        // they appear in the XML notification.
        for (int i = 0; i < keys.length; i++) {
            if (attributeQuantityName.equals(keys[i])) {
                key = (String)keys[i];
            }
        }
        
        if (key == null) {
            // If this is the first time we have seen this tag, then set the
            // default quantity value to one.
            vector = new Vector();
            vector.addElement(new Integer(1));
            quantValues.put(attributeQuantityName,vector);
        } else {
            // Add one to the quantity to include this tag in the count.
            vector= (Vector)quantValues.get(key);
            Integer value = (Integer)vector.elementAt(0);
            vector.setElementAt(new Integer(value.intValue()+1),0);
            quantValues.put(key,vector);
        }
    }
    
    /**
     * This method will check to see if any of the user rules given to the
     * processor can be applied to the current XML tag.
     *
     * @param node The Node that represents the XML tag in the DOM.
     * @return A string containing either a Siena attribute name or a processing
     *         instruction.
     */
    
    private String tryToApplyUserRules(Node node) {
        Vector tempVector = null;
        int index = 0;
        int ruleCounter = 0;            // This is used to count the number of
        // rules that could have been applied to
        // to the node. If more than one a log
        // message is generated.
        StringBuffer buffer = null;
        
        // The error buffer is used to store the warning messages when two user
        // rules can be applied to the same node.
        
        StringBuffer errorBuffer = new StringBuffer("WARNING: \n");
        String attribute = null;
        Rule rule = null;
        
        tempVector = (Vector)userRules.get(node.getLocalName());
        errorBuffer.append("Multiple user rules could be applied to a single XML tag.");
        errorBuffer.append("\nXML Tag: ");
        errorBuffer.append(node.getLocalName());
        
        if (tempVector != null) {
            for (int i = 0; i < tempVector.size(); i++) {
                rule = (Rule)tempVector.elementAt(i);
                // If the rule can be applied to the XML tag then apply it!
                if (rule.doesRuleApplyToNode(node)) {
                    ruleCounter++;
                    attribute = rule.apply(node);
                    errorBuffer.append("\nRule: ");
                    errorBuffer.append(rule.toString());
                }
            }
        }
        
        // To make the rest of the applyRule code to work for user rules
        // that have 'unique' actions, a slash has to be inserted before the
        // dollar sign otherwise the siena attribute name will not come
        // out correctly.
        
        if (attribute != null) {
            index = attribute.indexOf("$");
            if (index != -1) {
                buffer = new StringBuffer();
                buffer.append(attribute.substring(0,index));
                buffer.append("/");
                buffer.append(attribute.substring(index));
                attribute = buffer.toString();
            }
        }
        
        // If more than one user rule could have been applied to the node then
        // print out the warning message. The message is either logged to the
        // log file or printed out to the Error stream.
        if (ruleCounter > 1) {
            errorBuffer.append("\nNote: Last rule to match is the one applied to the tag.\n\n");
            if(logFile != null) {
                try {
                    logFile.write(errorBuffer.toString());
                    logFile.flush();
                } catch (java.io.IOException exception) {
                    System.err.println("Unable to write to log file: "+logfileName);
                    System.err.println("Exception: "+exception.getMessage());
                }
            } else {
                System.err.println(errorBuffer.toString());
            }
        }
        
        return attribute;
    }
    
    /**
     * This method sets the rules that the user wants to be used.
     *
     * @param filename Name of the file where the user rules are located.
     * @throws SienaException If the user rules file is uanble to be opened or parse this exception is thrown.
     */
    
    public void setUserRules(String filename) throws siena.SienaException {
        java.io.FileInputStream input = null;
        RulesGenerator generator = null;
        
        try {
            input = new java.io.FileInputStream(filename);
            generator = new RulesGenerator(input);
        } catch (java.io.IOException exception) {
            throw new siena.SienaException(exception.getMessage());
        }
        userRules = generator.getMap();
    }
    
    /**
     * This method will clear the user rules
     */
    
    public void clearUserRules() {
        userRules = null;
    }
    
    /**
     * This is an inner class used to test out the private methods of this class.
     *
     * @author Chris Corliss
     */
    
    public static class TEST extends TestCase {
        private XMLProcessor processor = null;
        private ElementImpl testNode = null;
        private CoreDocumentImpl document = null;
        private Set keySet = null;
        private Rule testRule = null;
        
        public TEST(String testName) {
            super(testName);
        }
        
        public static Test suite() {
            return ( new TestSuite(TEST.class));
        }
        
        protected void setUp() throws Exception {
            super.setUp();
            processor = new XMLProcessor();
            document = new CoreDocumentImpl();
        }
        
        public void testPopulateNotification() {
            String result = null;
            double value = 0;
            siena.Notification notification = new siena.Notification();
            Vector intVector = new Vector();
            Vector doubleVector = new Vector();
            Vector multipleStringVector = new Vector();
            
            System.out.println(" - Test: XMLProcessor - populateNotification method");
            
            intVector.addElement(new Integer(1));
            doubleVector.addElement(new Double(0.01));
            multipleStringVector.addElement("test");
            multipleStringVector.addElement("one");
            multipleStringVector.addElement("two");
            
            processor.attributeValue = new HashMap();
            processor.attributeValue.put("intTest",intVector);
            processor.attributeValue.put("doubleTest",doubleVector);
            processor.attributeValue.put("stringTest",multipleStringVector);
            
            processor.populateNotification(notification);
            
            Assert.assertEquals("The notification integer value was not correct.", 1, notification.getAttribute("intTest").intValue());
            
            value = notification.getAttribute("doubleTest").doubleValue();
            Assert.assertEquals("The notification double value was not correct.", .01, value, (0.1-value));
            
            result = notification.getAttribute("stringTest").toString();
            Assert.assertEquals("The value for the stringTest attribute is wrong.", "\"test, one, two\"", result);
        }
        
        public void testGenerateQuantityValue() {
            String name = null;
            Node parent = null;
            Node child1 = null;
            Node child2 = null;
            Node child3 = null;
            Vector values = null;
            int number = 0;
            
            System.out.println(" - Test: XMLProcessor - generateQuantityValue method");
            
            processor.attributeValue = new HashMap();
            
            parent = document.createElementNS("","Parent");
            child1 = document.createElementNS("","only");
            child2 = document.createElementNS("","second");
            child3 = document.createElementNS("","only");
            
            parent.appendChild(child1);
            processor.generateQuantityValue("only", child1);
            values = (Vector)processor.quantValues.get("only$quant");
            Assert.assertNotNull("The child \"only\" quantity attribute was not created.", values);
            number = ((Integer)values.elementAt(0)).intValue();
            Assert.assertEquals("The quantity for \"only\" child is incorrect.", 1, number);
            
            parent.appendChild(child2);
            processor.generateQuantityValue("second", child2);
            values = (Vector)processor.quantValues.get("second$quant");
            Assert.assertNotNull("The child \"second\" quantity attribute was not created.", values);
            number = ((Integer)values.elementAt(0)).intValue();
            Assert.assertEquals("The quantity for \"second\" child is incorrect.", 1, number);
            
            parent.appendChild(child3);
            processor.generateQuantityValue("only", child1);
            values = (Vector)processor.quantValues.get("only$quant");
            Assert.assertNotNull("The child \"only\" quantity attribute was not created with two siblings.", values);
            number = ((Integer)values.elementAt(0)).intValue();
            Assert.assertEquals("The quantity for \"only\" child is incorrect with two siblings.", 2, number);
        }
        
        public void testApplyRulesForNoAttributesGenerated() {
            System.out.println(" - Test: XMLProcessor - applyRules method, no attributes generated from node");
            
            processor.applyRules(document.createElementNS("","Test"),"");
            Assert.assertTrue("No SIENA attributes were to be expected but there "+
            "were some.", processor.attributeValue.isEmpty());
            processor.applyRules(document.createTextNode("TEST"),"");
            Assert.assertTrue("No SIENA attributes were to be expected but there "+
            "were some.", processor.attributeValue.isEmpty());
        }
        
        public void testApplyRulesAndUniqueProcessingInstruction() {
            int intValue = 0;
            double doubleValue = 0;
            Vector values = null;
            String value = null;
            org.apache.xerces.dom.ElementImpl parent = (ElementImpl)document.createElementNS("","parent");
            Node child = document.createElementNS("","child");
            Node child2 = document.createElementNS("","child2");
            
            System.out.println(" - Test: XMLProcessor - applyRules method, unique processing instruction encountered");
            
            processor.attributeValue = new HashMap();
            
            processor.xmlRules = new HashMap();
            processor.xmlRules.put("parent","parent/$name");
            
            parent.setAttribute("name","true");
            child.appendChild(document.createTextNode("passed"));
            parent.appendChild(child);
            processor.applyRules(parent,"");
            values = (Vector)processor.attributeValue.get("parent/true/child");
            if (values == null) {
                fail("No values for the attribute parent/true/child.");
            }
            value = (String)values.elementAt(0);
            Assert.assertNotNull("applyRules did not generate the correct SIENA attribute for unique instruction. "+
            "Expecting parent/true/child but got null.", value);
            Assert.assertEquals("The value for the partent/true/child is not correct. ", "passed",value);
            
            processor.xmlRules.put("parent","parent/$child");
            processor.applyRules(parent,"");
            values = (Vector)processor.attributeValue.get("parent/passed/child");
            if (values == null) {
                fail("No values for the attribute parent/passed/child.");
            }
            value = (String)values.elementAt(0);
            Assert.assertNotNull("applyRules did not generate the correct SIENA attribute for unique instruction.", value);
            Assert.assertEquals("The value for the partent/passed/child is not correct. ","passed",value);
            
        }
        
        public void testApplyRulesProcessingInstructions() {
            org.apache.xerces.dom.ElementImpl parent = (ElementImpl)document.createElementNS("","parent");
            Node child = document.createElementNS("","child");
            
            System.out.println(" - Test: XMLProcessor - applyRules method, processing instructions encountered (IGNORE and PATH)");
            
            child.appendChild(document.createTextNode("test"));
            parent.appendChild(child);
            
            processor.xmlRules.put("parent","IGNORE");
            processor.attributeValue = new HashMap();
            processor.applyRules(parent,"");
            Assert.assertNull("IGNORE processing instruction was not executed.", processor.attributeValue.get("parent"));
            
            processor.xmlRules.put("parent","PATH");
            processor.applyRules(parent,"");
            Assert.assertNotNull("PATH processing instruction was not executed.", processor.attributeValue.get("parent/child"));
            processor.applyRules(parent,"prefix/");
            Assert.assertNotNull("Prefix parameter processing not correct.", processor.attributeValue.get("prefix/parent/child"));
        }
        
        public void testApplyRulesNumericalData() {
            int intValue = 0;
            double doubleValue = 0;
            Vector values = null;
            String value = null;
            org.apache.xerces.dom.ElementImpl parent = (ElementImpl)document.createElementNS("","parent");
            Node child = document.createElementNS("","child");
            Node child2 = document.createElementNS("","child2");
            
            System.out.println(" - Test: XMLProcessor - applyRules method, node contains numerical data");
            
            child.appendChild(document.createTextNode("1"));
            child2.appendChild(document.createTextNode("100"));
            parent.appendChild(child);
            parent.appendChild(child2);
            processor.xmlRules = new HashMap();
            processor.xmlRules.put("child","low");
            processor.xmlRules.put("child2","high");
            processor.applyRules(parent,"");
            values = (Vector)processor.attributeValue.get("low");
            intValue = ((Integer)values.elementAt(0)).intValue();
            Assert.assertEquals("Low integer value not correct.",1,intValue);
            values = (Vector)processor.attributeValue.get("high");
            intValue = ((Integer)values.elementAt(0)).intValue();
            Assert.assertEquals("High integer value not correct.",100,intValue);
            
            parent = (org.apache.xerces.dom.ElementImpl)document.createElementNS("","Parent");
            child = document.createElementNS("","child");
            child2 = document.createElementNS("","child2");
            child.appendChild(document.createTextNode("0.1"));
            child2.appendChild(document.createTextNode("100.001"));
            parent.appendChild(child);
            parent.appendChild(child2);
            
            processor.xmlRules = new HashMap();
            processor.xmlRules.put("child","low");
            processor.xmlRules.put("child2","high");
            processor.attributeValue = new HashMap();
            
            processor.applyRules(parent,"");
            
            values = (Vector)processor.attributeValue.get("low");
            doubleValue = ((Double)values.elementAt(0)).doubleValue();
            Assert.assertEquals("Low integer value not correct.",.1,doubleValue,(.1-doubleValue));
            values = (Vector)processor.attributeValue.get("high");
            doubleValue = ((Double)values.elementAt(0)).doubleValue();
            Assert.assertEquals("High integer value not correct.",100.001,doubleValue,(100.001-doubleValue));
            
            processor.xmlRules = new HashMap();
            processor.attributeValue = new HashMap();
        }
        
        public void testGenerateMinMaxValuesWithIntegers() {
            siena.Notification notification = new siena.Notification();
            MinMax temp = new MinMax();
            int value = 0;
            
            System.out.println(" - Test: XMLProcessor - generateMinMaxValues method, using integers");
            
            temp.adjust(-100);
            temp.adjust(100);
            processor.minMaxValues = new HashMap();
            processor.minMaxValues.put("test",temp);
            processor.generateMinMaxValues(notification);
            value = notification.getAttribute("test$min").intValue();
            Assert.assertEquals("Min is not the correct value.", -100,value);
            value = notification.getAttribute("test$max").intValue();
            Assert.assertEquals("Max is not the correct value.", 100,value);
            Assert.assertTrue("minMaxValues HashMap was not cleared at the end.", processor.minMaxValues.isEmpty());
        }
        
        public void testGenerateMinMaxValuesWithDoubles() {
            siena.Notification notification = new siena.Notification();
            MinMax temp = new MinMax();
            double value = 0;
            
            System.out.println(" - Test: XMLProcessor - generateMinMaxValues method, using doubles");
            
            temp.adjust(-100.00);
            temp.adjust(100.00);
            processor.minMaxValues = new HashMap();
            processor.minMaxValues.put("test",temp);
            processor.generateMinMaxValues(notification);
            value = notification.getAttribute("test$min").doubleValue();
            Assert.assertEquals("Min is not the correct value. Expectin -100.00."+value, -100.00,value, (-100.00-value));
            value = notification.getAttribute("test$max").doubleValue();
            Assert.assertEquals("Max is not the correct value. Expecting 100.00."+value, 100.00,value, (100-value));
        }
        
        public void testTryToApplyUserRules() {
            Node testNode = document.createElementNS("","testNode");
            Rule testRule = new Rule("testNode","passed");
            testRule.addCondition(new siena.xml.rules.ExistsCondition(true));
            String result = null;
            
            System.out.println(" - Test: XMLProcessor - tryToApplyUserRules method");
            
            processor.userRules = new HashMap();
            Vector rules = new Vector();
            rules.addElement(testRule);
            processor.userRules.put("testNode",rules);
            
            testNode.appendChild(document.createTextNode("data"));
            result = processor.tryToApplyUserRules(testNode);
            Assert.assertEquals("Did not get the right result from the user rules.","passed",result);
            
            testRule = new Rule("testNode","unique:passed");
            testRule.addCondition(new siena.xml.rules.ExistsCondition(true));
            rules.removeElementAt(0);
            rules.insertElementAt(testRule,0);
            processor.userRules.put("testNode", rules);
            
            result = processor.tryToApplyUserRules(testNode);
            Assert.assertEquals("Did not get the right result from the user rules.","passed/$testNode",result);
        }
        
        /**
         * Main method to execution to run the tests
         *
         * @param array of strings passed in from the command line
         */
        
        public static void main(String[] args) {
            junit.textui.TestRunner.run(suite());
        }
    }
}
