//
//  This file is part of SienaXML, an XML interface to the 
//  Siena wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Author: Chris Corliss <corliss@falcon.stortek.com>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2002 University of Colorado
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: XMLSubscriptionHandler.java,v 1.1.1.1 2002/06/07 22:42:37 jgiacomo Exp $
//
package siena.xml;

import siena.AttributeValue;
import siena.AttributeConstraint;
import siena.Op;
import siena.Filter;
import siena.SienaException;
import siena.Notification;

import java.util.zip.GZIPInputStream;
import java.util.Vector;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import siena.Notifiable;
import siena.xml.xpath.*;
import siena.xml.rules.Rule;

import junit.framework.*;

/**
 * This class handles the processing of an XPath statement and runging it into a set of filters that can be used as a subscription to a XML notification. This class also makes sure that any XML notification that is recieved from a server does match the XPath that was used to generate the subscription.
 * @author Chris Corliss
 * @version 1 
 */
public class XMLSubscriptionHandler implements siena.Notifiable, Subject {

    // To prevent this class being created each time a handler is created I amde it
    // static. This makes sense as the rules will not change, and if so all
    // hanlders should know about it.


    /**
     * This is the object that will process the XSchema and create the
     * correct mapping from an XML tag into a SIENA attribute. 
     */
    private static XSchemaProcessor processor = new XSchemaProcessor();

    /**
     * This is a map to go from an XML tag into a SIENA attribute name. 
     */
    private HashMap rules = null;

    /**
     * A wrapper around the XML DOM parser. The parser is used when evaluating 
     * the XPath against the XML notification received from the SIENA server.
     */
    private DOMParserWrapper parser;

    /**
     * This variable holds the XPath that was passed in as an arguement to the 
     * buildFilters method. 
     */
    private XPath xpath;
    
    /**
     * A vector containing all of the observers to this class. The obeservers 
     * are applications that are to be notified when a XML notification has 
     * arrived that matches the XPath.
     *
     * @associates <{Observer}>
     */

    private Vector observersVector = new java.util.Vector();

    /** This variable holds the XML Notification which was returned by the 
     * SIENA servers as matching the subscription(s) generated by this object.
     */    
    private String XMLNotification;

    /** This is the Object that should be notified when a XML Notification has 
     * been returned.
     */    
    private Notifiable notifiable;

    /** This contains a list of predicate functions that can not be converted 
     * to SIENA name attributes at this time.
     */    
    private HashMap unhandable;

    /**
     * This is a vector that holds all the filters that have been generated when
     * processing the XPath 
     */
    
    private Vector filters = new Vector();

    /** This is the constructor for this object.
     * @param notifiable The object to be notified when an XML notification is 
     *                   returned by the SIENA servers.
     * @throws IOException If there is a problem in opening the default rules 
     *                     file this exception is thrown.
     */    
    public XMLSubscriptionHandler(Notifiable notifiable) throws java.io.IOException {

        // We need a dummy object to put into the hash map. null can not be used
        // since some methods check for null to see if a entry in the hash map
        // exists or not.
        Object dummy = new Object();

        this.notifiable = notifiable;
        parser = new DOMParserWrapper();

        unhandable = new HashMap();
        unhandable.put("translate", dummy);
        unhandable.put("count", dummy);
        unhandable.put("id", dummy);
        unhandable.put("local-name", dummy);
        unhandable.put("namespace-uri", dummy);
        unhandable.put("name", dummy);
        unhandable.put("normalize-space", dummy);
        unhandable.put("not", dummy);
        unhandable.put("true", dummy);
        unhandable.put("false", dummy);
        unhandable.put("lang", dummy);
        unhandable.put("sum", dummy);
        unhandable.put("floor", dummy);
        unhandable.put("ceiling", dummy);
        unhandable.put("round", dummy);
    }
    
    /** This function is used to notify this object of an SIEAN notification 
     * that matches the SIENA subscriptions generated by this command.
     * @param notification The notification that was returned by the 
     *                     SIENA servers.
     * @throws SienaException If there is a problem extracting the XML 
     *                        notification from the SIENA notification 
     *                        this exception is thrown.
     */    
    public void notify(Notification notification) throws SienaException {
        java.io.ByteArrayInputStream xmlbytes = null;
        String xmlString = null;
        org.w3c.dom.NodeList nodes = null;
        org.w3c.dom.Node root = null;
        org.apache.xpath.objects.XObject result = null;

        xmlbytes = new java.io.ByteArrayInputStream(notification.getAttribute("_xml").byteArrayValue());

        try {
            java.io.ObjectInputStream xmlStream = new java.io.ObjectInputStream(new GZIPInputStream(xmlbytes));
            xmlString = xmlStream.readUTF();
        } catch (Exception exception) {
            throw new siena.SienaException("Unable to get the XML Notification from the Siena Notification");
        }
        
        // Execute the XPath against the notification. If it returns at least
        // one node then the notification should be returned otherwise disgard
        // the notification and wait for the next one.
        try {
            root = parser.parse(xmlString);
            result = org.apache.xpath.XPathAPI.eval(root,xpath.toString());
            nodes = result.nodelist();
        } catch (javax.xml.transform.TransformerException exception) {
            throw new siena.SienaException("Trouble executing the xpath subscription against the XML notification");
        }

        if (nodes.getLength() > 0) {
            XMLNotification = xmlString;
            this.inform();
        }
    }
    
    /** This function is used to notify this object of an SIEAN notification that 
     *  matches the SIENA subscriptions generated by this command.
     * @param s An array of subscriptions to handle
     * @throws SienaException If there is a problem extracting the XML 
     *                        notification from the SIENA notification this 
     *                        exception is thrown.
     */    
    public void notify(Notification[] s) throws SienaException {
        for (int i = 0; i < s.length; i++) {
            this.notify(s[i]);
        }
    }
    
    /** Attach an observer to this object
     * @param observer The object to include in the list of observing objects.
     */    
    public void attach(Observer observer){
        observersVector.addElement(observer);
    }
    
    /** Remove an observer from the list of observers.
     * @param observer The object/observer to remove from the list.
     */    
    public void detach(Observer observer){
        observersVector.removeElement(observer);
    }
    
    /** Used to inform observers of a XML notification that was recieved from 
     * the SEINA servers.
     */    
    public void inform(){
        java.util.Enumeration enumeration = observers();
        while (enumeration.hasMoreElements()) {
            ((Observer)enumeration.nextElement()).update(this);
        }
    }
    
    /** Return a enumeration of observers for this object
     * @return Enumeration object containing the observers for this object.
     */    
    public Enumeration observers(){
        return ((java.util.Vector) observersVector.clone()).elements();
    }
    
    /** Returns the XML notification received from the SIEAN servers.
     * @return String containing the XML notification.
     */    
    public String getNotification() {
        return XMLNotification;
    }
    
    /** Returns the notifiable object
     * @return The Notifiable object
     */    
    public Notifiable getNotifiable() {
        return notifiable;
    }
    
    /**
     * This method processes the XPath expression passed in and returns a SIENA
     * filter that will correctly subscribe to the events that are generated from
     * an XML Notification based on the XSchema given to this object.
     *
     * @return Array of filters to be used as subscriptions.
     * @param subscription The subscription, or XPath, statement to be converted 
     *                     into filters.
     * @throws SienaException If no filters can be created this exception is thrown.
     */
    
    public Filter[] buildFilter(String subscription) throws siena.SienaException {
        Step step = null;
        String stepName = null;
        String newprefix = "";
        Predicate predicate = null;
        boolean isXPathValid = true;
        int indexOfParen = -1;
        String prefix = "";
        String stepRule = null;
        CompoundPredicate compoundPredicate = null;
        Filter filter = null;
        BinaryNode root = null;
        int maxFilters = 0;
       

        if (xpath != null && subscription.equals(xpath.toString()) && filters != null) {
            // For some reason Java did not like the following line:
            // return (Filter[])filters.toArray()
            // Exception thrown was ClassCastExceptionL [Ljava.lang.Object;
            // Therefore this ugly code.
            Filter[] filterArray = new Filter[filters.size()];
            for (int i = 0; i < filterArray.length; i++) {
                filterArray[i] = (Filter)filters.elementAt(i);
            }
	    return filterArray;
        }
        
        this.xpath = new XPath(subscription);
        
        // break the path into its individual steps
        
        for (int i = 0; i < xpath.getNumberOfSteps(); i++) {

            step = xpath.getStepAt(i);
            stepName = step.getNodeTest();

            if (step.isNodeTestAFunction()) {
                indexOfParen = stepName.indexOf("(");
                stepName = stepName.substring(0,indexOfParen);
            }
            
            /**
             * If the users is trying to use the '*' node test then go to the next
             * step because it is to difficult to figure out a filter with a "star"
             * node test.
             */
            
            if (stepName.equals("*")) {
                continue;
            }
            
            stepRule = (String)rules.get(stepName);
            
            if (stepRule == null) {
                stepName = null;
            } else if (stepRule.equals("PATH")) {
                newprefix += step.getNodeTest()+"/";                
            }
            
            /**
             * Subscriptions are mainly created from predicates and not the
             * node tedt or axis part of the Step.
             */
            if (step.hasPredicate()) {
                predicate = step.getPredicate();
                /**
                 * If the predicate type is a composite type then there are
                 * multiple predicates that need to be handled. 
                 */
                if (predicate.getType() == Predicate.COMPOUND) {
                    compoundPredicate = (CompoundPredicate)predicate;
                    root = compoundPredicate.getTree();
                    isXPathValid &= this.buildFiltersFromTree(root, newprefix, stepRule, 0);
                    filters.addAll((Vector)root.getData());
                } else {
                    filter = new Filter();
                    isXPathValid &= addConstraint(predicate, newprefix, stepRule,filter);
                    filters.addElement(filter);
                }
            } else if (stepRule != null) {
                filter = new Filter();                
                filter.addConstraint(stepName, new AttributeConstraint(Op.ANY, "any"));
                filters.addElement(filter);
                isXPathValid = true;
            }
            prefix = newprefix;
        }
         
        if (isXPathValid) {
            // For some reason Java did not like the following line:
            // return (Filter[])filters.toArray()
            // Exception thrown was ClassCastExceptionL [Ljava.lang.Object;
            // Therefore this ugly code.
            Filter[] filterArray = new Filter[filters.size()];
            for (int i = 0; i < filterArray.length; i++) {
                filterArray[i] = (Filter)filters.elementAt(i);
            }
	    return filterArray;
        } else {
            throw new siena.SienaException("Invalid Xpath for subscription: "+
                    "unable to determine what notification attributes to "+
                    "subscribe to.");
        }
    }
    
    /** This is used to build filters from a compound predicate tree.
     * @param node The current node in the tree to be processed
     * @param prefix Any prefix to be applied to the SIENA attribute name
     * @param currentName Passed to the addConstraint method.
     * @param currentFilter The filter for the addConstraint method to manipulate.
     * @return True if filters were created, otherwise false.
     */    
    private boolean buildFiltersFromTree(BinaryNode node, String prefix, String currentName, int currentFilter) {
        boolean result = false;
        Object data = node.getData();
        Vector nodeFilters = new Vector();
        Vector leftFilters = null;
        Vector rightFilters = null;
                
        // If this node is not a leaf process its children first.
        if (!node.isLeaf()) {

            // Left child processing
            result |= buildFiltersFromTree(node.getLeftChild(), prefix, currentName, currentFilter);                

            // Right child processing.
            result |= buildFiltersFromTree(node.getRightChild(), prefix, currentName, currentFilter);

            // Get the filters from the children
            leftFilters = (Vector)node.getLeftChild().getData();
            rightFilters = (Vector)node.getRightChild().getData();
            
            // Based on the NODE type combine the children filters into one
            // set of filters.

            if (((String)data).equals("AND")) {
                // AND Nodes have to create all the permutations of the
                // children nodes
                for (int i = 0; i < leftFilters.size(); i++) {
                    for (int j = 0; j < rightFilters.size(); j++) {
                        nodeFilters.addElement(addFilters((Filter)leftFilters.elementAt(i), (Filter)rightFilters.elementAt(j)));
                    }
                }
            } else {
                // OR nodes only have to add the filters together
                for (int i = 0; i < leftFilters.size(); i++) {
                    nodeFilters.addElement(leftFilters.elementAt(i));;
                }                
                for (int i = 0; i < rightFilters.size(); i++) {
                    nodeFilters.addElement(rightFilters.elementAt(i));
                }
            }
            node.setData(nodeFilters);
        } else {
            // Processing for Leaf nodes. Simply create the filter for this
            // predicate.
            Filter filter = new Filter();
            result |= addConstraint((Predicate)data, prefix, currentName, filter);
            nodeFilters.addElement(filter);
            node.setData(nodeFilters);
        }        
        return result;
    }
    
    /** This method will add the contents of the two fitlers together and 
     * return a new filter.
     * @param f1 Filter object to use
     * @param f2 Filter object to use
     * @return New Filter containing the constraints of f1 and f2
     */    
    private Filter addFilters(Filter f1, Filter f2) {
        Filter newFilter = new Filter(f2);
        Iterator iterator = null;
        String name = null;
        for(iterator = f1.constraintNamesIterator(); iterator.hasNext();) {
            name = (String)iterator.next();
            for(Iterator li = f1.constraintsIterator(name); li.hasNext();) {
                newFilter.addConstraint(name, (AttributeConstraint)li.next());
            }
        }        
        return newFilter;
    }
    
    /** This method will add constraint(s) to the filter based on the 
     * predicate passed in.
     * @param predicate Predicate to use to generate the attribute constraints.
     * @param prefix Prefix to append to the SIENA attribute name
     * @param currentStepName Current node test name of the XPath step being looked at
     * @param filter The Filter object to add the constraint to.
     * @return True if an attribute constraint was added to the filter, otherwise 
     *         false.
     */    
    private boolean addConstraint(Predicate predicate, String prefix, String currentStepName, Filter filter) {
        String currentPrefix = prefix;
        String notificationAttribute = null;
        String arg = null;
        String operation = null;
        int index = 0;
        boolean status = false;

        AttributeConstraint constraint = null;
        EqualityPredicate equalPredicate = null;
        PositionPredicate positionPredicate = null;
        ComparisonPredicate comparePredicate = null;
        FunctionPredicate functionPredicate = null;
        
        // Is the current predicate an equality expression?
        if (predicate.getType() == Predicate.EQUALITY) {
            equalPredicate = (EqualityPredicate)predicate;
            notificationAttribute = (String)rules.get(equalPredicate.getTarget().replace('@',' ').trim());
            if (equalPredicate.isIntegerComparison()) {
                constraint = new AttributeConstraint(Op.EQ, equalPredicate.getIntValue());
            } else if (equalPredicate.isDoubleComparison()) {
                constraint = new AttributeConstraint(Op.EQ, equalPredicate.getDoubleValue());
            } else {
                constraint = new AttributeConstraint(Op.SS, equalPredicate.getStringValue());
            }
            // Is the current predicate using the position of the node?
        } else if (predicate.getType() == Predicate.POSITION) {
            positionPredicate = (PositionPredicate)predicate;
            
            if (currentStepName == null) {
                return false;
            }
            
            index = currentStepName.indexOf("$");
            if (index != -1) {
                notificationAttribute = currentStepName.substring(0,index-1).trim()+"$quant";
            } else {
                notificationAttribute = currentStepName;
            }
            
            index = positionPredicate.getPosition();
            if (index == -1 || index == 1) {
                constraint = new AttributeConstraint(Op.ANY, "any");                
            } else {
                constraint = new AttributeConstraint(Op.GE, positionPredicate.getPosition());
            }
        } else if (predicate.getType() == Predicate.COMPARE) {

            comparePredicate = (ComparisonPredicate)predicate;

            operation = comparePredicate.getOp();
            index = operation.indexOf(">");
            if (index != -1) {
                notificationAttribute = comparePredicate.getStringArg()+"$max";
            } else {
                notificationAttribute = comparePredicate.getStringArg()+"$min";
            }

            // Check to see if the predicate is comparing an integer or a double
            if (comparePredicate.isThereAnIntegerArg()) {
                constraint = new AttributeConstraint(Op.op(operation), comparePredicate.getIntegerArg());
            } else {
                constraint = new AttributeConstraint(Op.op(operation), comparePredicate.getDoubleArg());
            }
            
            // Because of the way the notification is setup the prefix must be cleared 
            // in order to match the notification attribute.
            currentPrefix="";
        } else {
            functionPredicate = (FunctionPredicate)predicate;
            operation = functionPredicate.getFunctionName();            
            if (unhandable.get(operation) != null) {
                return false;
            }
            if (functionPredicate.isPredicateStringFunction()) {
                for (int i = 0; i < functionPredicate.getNumberOfArgs(); i++) {

                    /**
                     * Unquoted strings are candiates for notification attribute
                     * names.
                     */

                    if (!functionPredicate.isArgAQuotedString(i)) {

                        /**
                         * Get the value of the first arg and strip it of 
                         * any etreneous characters.
                         */

                        arg = functionPredicate.getStringArgAt(i).replace('@',' ').trim();
                        /**
                         * Get the rule for the value.
                         */

                        notificationAttribute = (String)rules.get(arg);

                        // There was a rule for this value so continue
                        // processing.

                        if (notificationAttribute != null) {

                            // If the function was a substring function or the
                            // contains function then check to see if the
                            // other arguement was a quoted string. If so, do
                            // a substring operation on the notification
                            // attribute using the value of the arguement that
                            // was quoted as the operands in the constraint..

                            int test = operation.indexOf("substring"); 
                            test += operation.indexOf("contains"); 
                            if (test >= -1 && functionPredicate.isArgAQuotedString((i+1)%2)) {
                                constraint = new AttributeConstraint(Op.SS, 
                                        functionPredicate.getStringArgAt((i+1)%2));
                                break;
                            } else {
                                constraint = new AttributeConstraint(Op.ANY,
                                        "any");

                                filter.addConstraint(notificationAttribute, 
                                        constraint);
                                status = true;
                            }
                        }
                    }
                    notificationAttribute = null;
                }
            }
        }

        if (notificationAttribute != null) {
            filter.addConstraint(currentPrefix+notificationAttribute, constraint);
            status = true;
        }
        return status;
    }

    /** This sets the style sheet for the handler.
     * @param filename Name of the file containing the XSchema
     * @throws SienaException If there is a problem opening or parsing the file 
     *                        this exception is thrown.
     */    
    public void setStylesheet(String filename) throws siena.SienaException {
        try {
            processor.setXSchema(filename);
        } catch (Exception exception) {
            throw new siena.SienaException(exception.getMessage());
        }
        rules = processor.getXMLRules();
    }

    /** White box testing class for XMLSubscriptionHandler
     */    
    public static class TEST extends TestCase implements siena.Notifiable {
        XMLSubscriptionHandler subHandler = null;

        public TEST(String name) throws java.io.IOException {
            super(name);
        }

        public static Test suite() {
            return ( new TestSuite(TEST.class));
        }
        
        protected void setUp() throws Exception {
            super.setUp();
            subHandler = new XMLSubscriptionHandler(this);
        }

        public static void main(String[] args) {
            junit.textui.TestRunner.run(suite());
        }              

        public void testAddConstraint() {
            Filter filter = new Filter();
            System.out.println(" - Test: XMLSubscriptionHandler - addConstraint method");
            Predicate eqPredicate = new EqualityPredicate("test","1");
            Predicate posPredicate1 = new PositionPredicate(1);
            Predicate posPredicate2 = new PositionPredicate("last()");
            Predicate posPredicate3 = new PositionPredicate("position() = 1");            
            Predicate compPredicate1 = new ComparisonPredicate("test > 1");
            Predicate compPredicate2 = new ComparisonPredicate("test < 1");
            Predicate compPredicate3 = new ComparisonPredicate("test >= 1");
            Predicate compPredicate4 = new ComparisonPredicate("test <= 1");            
            Predicate funcPredicate = new FunctionPredicate("substring(name,\"test\")");
            
            subHandler.rules = new HashMap();            
            subHandler.rules.put("test","passed");
            
            Assert.assertTrue("addConstraint failed with an EqeualityPredicate.", subHandler.addConstraint(eqPredicate,"","",filter));
            Assert.assertTrue("The equality constraint was not added to the filter.",filter.containsConstraint("passed"));
            Assert.assertTrue("addConstraint failed with PositionPredicate(1).", subHandler.addConstraint(posPredicate1,"","step",filter));
            Assert.assertTrue("The position constraint:1 was not added to the filter.",filter.containsConstraint("passed"));
            Assert.assertTrue("addConstraint failed with PositionPredicate(last())",subHandler.addConstraint(posPredicate2,"","step",filter));
            Assert.assertTrue("The position constraint:last() was not added to the filter.",filter.containsConstraint("passed"));
            Assert.assertTrue("addConstraint failed with PositionPredicate(position() = 1)",subHandler.addConstraint(posPredicate3,"","step",filter));            
            Assert.assertTrue("The position constraint:position() = 1 was not added to the filter.",filter.containsConstraint("passed"));
            Assert.assertTrue("addConstraint added a constraint for a PositionPredicate with a null currentStepName when it should not have.",!subHandler.addConstraint(posPredicate1,"",null,filter));
            Assert.assertTrue("addConstraint failed with ComparisionPredicate(test > 1)",subHandler.addConstraint(compPredicate1,"","",filter));
            Assert.assertTrue("The comparison constraint:test > 1 was not added to the filter.",filter.containsConstraint("passed"));
            Assert.assertTrue("addConstraint failed with ComparisionPredicate(test < 1)",subHandler.addConstraint(compPredicate2,"","",filter));
            Assert.assertTrue("The comparison constraint:test < 1 was not added to the filter.",filter.containsConstraint("passed"));
            Assert.assertTrue("addConstraint failed with ComparisionPredicate(test >= 1)",subHandler.addConstraint(compPredicate3,"","",filter));
            Assert.assertTrue("The comparison constraint:test >= 1 was not added to the filter.",filter.containsConstraint("passed"));
            Assert.assertTrue("addConstraint failed with ComparisionPredicate(test <= 1)",subHandler.addConstraint(compPredicate4,"","",filter));            
            Assert.assertTrue("The comparison constraint:test <= 1 was not added to the filter.",filter.containsConstraint("passed"));
            Assert.assertTrue("addConstraint failed with FunctionPredicate(substring(name, rob)) when there was no rule for name.",!subHandler.addConstraint(funcPredicate,"","",filter));
            subHandler.rules.put("name","passed");
            Assert.assertTrue("addConstraint failed with FunctionPredicate(substring(name, rob)).",subHandler.addConstraint(funcPredicate,"","",filter));
            Assert.assertTrue("The functino constraint:substring was not added to the filter.",filter.containsConstraint("passed"));
            Assert.assertTrue("addConstraint failed to add constraint with a prefix passed in.",subHandler.addConstraint(eqPredicate,"prefix/","",filter));
            Assert.assertTrue("The equality constraint with a prefix was not added to the filter.",filter.containsConstraint("prefix/passed"));
        }

        /**
         * This function is empty and does nothing since it is never called. It
         * is here to satisfy the requirement for the implemented interface.
         */
        public void notify(Notification[] s) throws SienaException {
        }
        
        /**
         * This function is empty and does nothing since it is never called. It
         * is here to satisfy the requirement for the implemented interface.
         */
        public void notify(Notification n) throws SienaException {
        }

    }
}
