//
//  This file is part of SienaXML, an XML interface to the 
//  Siena wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Author: Chris Corliss <corliss@falcon.stortek.com>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2002 University of Colorado
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: XSchemaProcessor.java,v 1.2 2002/06/20 23:55:42 jgiacomo Exp $
//
/*
 * XSchemaProcessor.java
 *
 * Created on November 23, 2001, 9:12 PM
 */

package siena.xml;

import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;

import java.util.Vector;
import java.util.HashMap;
import java.io.FileInputStream;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import java.io.IOException;
import siena.xml.rules.Rule;
import siena.xml.rules.RulesGenerator;

import junit.framework.*;
import org.apache.xerces.dom.ElementImpl;

/** This class is used to process an XSchema and produce a mapping of XML tags to SIENA attribute names.
 *
 * @author Chris Corliss
 * @version 0.1
 */
public class XSchemaProcessor {
    
    /**
     * This string contains the name of the file that contains the default rules
     * that are to be used when processing the XML Schema
     */
    
    private static final String DEFAULTRULES = "default.rules";
    
    /**
     * This string contains the name of the file that contains the rules for
     * XML tag values that are to be used when processing the XML Schema.
     */
    
    private static final String VALUERULES = "value.rules";
    
    /**
     * This map holds the default rules. It is generated by the RulesGenerator
     * object.
     *
     * @see RulesGenerator
     */
    
    private java.util.HashMap defaultRules = null;
    
    /** This map contains the most recent XML to SIENA map that was generated 
     * by this object.
     */    
    private java.util.HashMap currentRules = null;
    
    /**
     * This map holds the value rules read in from the file pointed to by the
     * VALUERULES attribute.
     */
    
    private java.util.HashMap valueRules = null;
    
    /**
     * This map holds all of the types defined by the XSchema that will not have
     * any data/value in them. Example: <bob></bob> is a XML tag that has no value
     * in it.
     */
    
    private java.util.HashMap noValueMap = null;
    
    /**
     * This is the parser to be used to parse the XML document and turn it
     * into a W3C DOM.
     * @since 11-05-2001
     */
    private DOMParserWrapper parser;
    
    /**
     * This string contains the name of the current type being defined in the
     * XSchema.
     */
    
    private String currentType = null;
    
    /** Creates new XSchemaProcessor */
    public XSchemaProcessor() {
        parser = new DOMParserWrapper();
        currentRules = new HashMap();
        noValueMap = new HashMap();
        
        try {
            java.io.InputStream defaultInput = this.getClass().getResourceAsStream(DEFAULTRULES);
            java.io.InputStream valueInput = this.getClass().getResourceAsStream(VALUERULES);
//            java.io.InputStream defaultInput = new FileInputStream(DEFAULTRULES);
//            java.io.InputStream valueInput = new FileInputStream(VALUERULES);
            RulesGenerator generator = new RulesGenerator(defaultInput);
            defaultRules = generator.getMap();
            generator = new RulesGenerator(valueInput);
            valueRules = generator.getMap();
        } catch (Exception exception) {
            System.err.println("Problem generatring the default rules. Aborting Program.");
            System.err.println("The exception caught is: "+exception);
            System.exit(1);
        }
    }
    
    /**
     * Specify the XSchema to be processed
     *
     * @param filename Name of the file that contains the XSchema
     * @throws IOException This is thrown if the file can not be opened.
     * @throws SAXException If there is a problem apply the default rules 
     *                      to the XSchema this exception is thrown.
     */
    
    public void setXSchema(String filename) throws IOException, SAXException {
        Node doc = parser.parse(new InputSource(new FileInputStream(filename)));
        generateValueInformation(doc.getFirstChild(),"");
        this.currentType = "";
        generateXMLRules(doc.getFirstChild(),"");
    }
    
    /**
     * Get the rules that are to be applied to the XML.
     *
     * @return HashMap containing a map between an XML tag and the rule action.
     */
    
    public HashMap getXMLRules() {
        return currentRules;
    }
    
    /** This method will generate a list of types defined by the XSchema that 
     * do not have a value associated with them. This means that these types only 
     * contain other XML tags.
     * @param node The node that represents an XML tag
     * @param currentType The current type being defined in the XSchema.
     */
    
    private void generateValueInformation(Node node, String currentType) {
        NodeList children = null;
        Vector rules = null;
        String nodeName = null;
        NamedNodeMap attributes = null;
        Node tempNode = null;
        String newCurrentType = null;
        
        rules = getRule(node, valueRules);
        nodeName = node.getLocalName();
        
        /**
         * This next section of code determines what the current type is that is
         * being processed. This is important when dealing with any rules that
         * specify the currentTag as the target. The current Tag is the same
         * as the name of the current type.
         */
        newCurrentType = "";
        
        /**
         * There can be the case where an element is declared first then
         * defined after it. In this case the complexType will not have a name .
         * attribute. To solve this we remember the current element name also.
         */
        
        if (nodeName != null && nodeName.equals("element")) {
            // If the node does not have any attributes then we can not get the
            // name of the element. Therefore say we are still in the previous
            // type.
            if (node.hasAttributes()) {
                attributes = node.getAttributes();
                tempNode = attributes.getNamedItem("name");
                if (tempNode != null) {
                    newCurrentType = tempNode.getNodeValue();
                }
            } else {
                newCurrentType = currentType;
            }
        }
        
        if (nodeName != null && nodeName.endsWith("Type")) {
            // If there are no attributes, or name attribute default the type to
            // the currentType passed in. This catches the problem where a
            // complexType is defined inside of the element using it.
            if (node.hasAttributes()) {
                attributes = node.getAttributes();
                tempNode = attributes.getNamedItem("name");
                if (tempNode != null) {
                    newCurrentType = tempNode.getNodeValue();
                } else {
                    newCurrentType = currentType;
                }
            } else {
                newCurrentType = currentType;
            }
        }
        
        /** Because applyRules needs the type of the current XML tag this line
         * sets the class's current type to the new current type varaible.
         */
        this.currentType = newCurrentType;
        /** End curentType Code **/
        
        if (rules != null) {
            useRules(rules, node);
        }
        
        // Process the children of this node.
        children = node.getChildNodes();
        for (int i = 0; i < children.getLength(); i++) {
            generateValueInformation(children.item(i), newCurrentType);
        }
    }
    
    /**
     * This method will generate the rules that are to be applied to an XML
     * document. This function is recursive.
     *
     * @param node The node that represents an XML tag
     * @param currentType The current type being defined in the XSchema.
     */
    
    private void generateXMLRules(Node node, String currentType) {
        boolean ruleHasBeenApplied = false;
        boolean skipXMLTag = false;
        Vector rules = null;
        String nodeName = null;
        String newCurrentType = null;
        NamedNodeMap attributes = node.getAttributes();
        Node tempNode = null;
        NodeList children = null;
        
        nodeName = node.getLocalName();
        
        /**
         * This next section of code determines what the current type is that is
         * being processed. This is important when dealing with any rules that
         * specify the currentTag as the target. The current Tag is the same
         * as the name of the current type. This code will also check to see if 
         * an element's type is a type that will not have a data value.
         */
        
        newCurrentType = "";
        
        /**
         * There can be the case where an element is declared first then
         * defined after it. In this case the complexType will not have a name .
         * attribute. To solve this we remember the current element name also.
         */
        
        if (nodeName != null && nodeName.equals("element")) {
            // If the node does not have any attributes then we can not get the
            // name of the element. Therefore say we are still in the previous
            // type.
            if (node.hasAttributes()) {
                tempNode = attributes.getNamedItem("name");
                if (tempNode != null) {
                    newCurrentType = tempNode.getNodeValue();
                }
                
                // Check to see if the type is one that has no value...                
                tempNode = attributes.getNamedItem("type");
                if (tempNode != null && noValueMap.get(tempNode.getNodeValue()) != null) {
                    skipXMLTag = true;
                }
                
            } else {
                newCurrentType = currentType;
            }
        }
        
        if (nodeName != null && nodeName.endsWith("Type")) {
            // If there are no attributes, or name attribute default the type to
            // the currentType passed in. This catches the problem where a
            // complexType is defined inside of the element using it.
            if (node.hasAttributes()) {
                tempNode = attributes.getNamedItem("name");
                if (tempNode != null) {
                    newCurrentType = tempNode.getNodeValue();
                } else {
                    newCurrentType = currentType;
                }
            } else {
                newCurrentType = currentType;
            }
        }
        
        this.currentType = currentType;
        /** End curentType Code **/

        
        ruleHasBeenApplied = false;
        
        // If we are not supposed to skip the XML tag....
        if (!skipXMLTag) {
            
            // Is there a rule for the XML Tag?
            rules = getRule(node, defaultRules);
            if (rules != null) {
                // Can the rule be applied?
                ruleHasBeenApplied = useRules(rules, node);
            }
            
            // If a specific rule could not be applied, see if a rule with the "any"
            // tag can be applied to the XML tag.
            if (!ruleHasBeenApplied) {
                rules = (Vector)defaultRules.get("any");
                useRules(rules, node);
            }
        } 
        
        // Process the children of this node.
        children = node.getChildNodes();
        for (int i = 0; i < children.getLength(); i++) {
            generateXMLRules(children.item(i), newCurrentType);
        }
    }
    
    /** This method will try to get a rule that can be applied to the current XML 
     * DOM node.
     * @param target Node the rule is going to be applied to
     * @param rules Set of rules to look through
     * @return Returns a vector of rule objects that can be applied to the XML Node.
     */    
    private Vector getRule(Node target, java.util.HashMap rules) {
        String key = null;
        String targetName = null;
        Vector rule = null;
        java.util.Set keys = null;
        java.util.Iterator iterator = null;
        
        if (target == null) {
            return null;
        }
        
        targetName = target.getLocalName();
        
        if (targetName == null) {
            return null;
        }
        
        rule = (Vector)rules.get(targetName);
        
        if (rule != null) {
            return rule;
        }
        
        keys = rules.keySet();
        iterator = keys.iterator();
        while (iterator.hasNext()) {
            key = (String)iterator.next();
            if (key.startsWith(targetName)) {
                return (Vector)rules.get(key);
            }
        }
        return null;
    }
    
    /** This method will apply the rules in the vector against the 
     * Node passed in. If a rule is applied to the node then the result is 
     * stored in the currentRules map.
     * @param rules Rules to be applied to the node
     * @param node XML DOM tree node to apply the rules to.
     * @return Returns true if a rule was applied to the node, otherwise false.
     */    
    private boolean useRules(Vector rules, Node node) {
        Rule rule = null;
        for (int i = 0; i < rules.size(); i++) {
            rule = (Rule)rules.elementAt(i);
            if (rule.doesRuleApplyToNode(node)) {
                try {
                    applyRule(rule, node);
                } catch (Exception ignore) {}
                return true;
            }
        }
        return false;
    }
    
    /** This method will apply a given Rule object to an XML DOM tree node.
     * @param rule The Rule object to apply to the node
     * @param node Target XML node for the rule to be applied to
     * @throws SienaException If the rule does not return any value after being applied to the node this exception is thrown.
     */    
    private void applyRule(Rule rule, Node node) throws siena.SienaException {
        String attributeName = null;
        String xmlTagName = null;
        String attribute = null;
        org.w3c.dom.NamedNodeMap attr = null;
        Node nameAttr = null;
        int index = -1;
        
        attributeName = rule.apply(node);
        
        if (attributeName == null) {
            throw new siena.SienaException("No rule action was returned!");
        }
        
        /**
         * If this node has been designated as not having a value, add it to the
         * noValueMap hash map and return from this function.
         */
        
        if (attributeName.equals("NOVALUE")) {
            noValueMap.put(this.currentType, "NOVALUE");
            return;
        }
        
        index = attributeName.indexOf('$');
        if (index != -1) {
            xmlTagName = attributeName.substring(0,index);
            attribute = attributeName.substring(index);
            if (xmlTagName.equals("currentTag")) {
                xmlTagName = currentType;
            }
            currentRules.put(xmlTagName, xmlTagName+"/"+attribute);
            return;
        }
        
        attr = node.getAttributes();
        nameAttr = attr.getNamedItem("name");
        if (nameAttr == null) {
            return;
        }
        currentRules.put(nameAttr.getNodeValue(),attributeName);
    }
    
    /** White Box test class for the XSchemaProcessor
     */    
    public static class TEST extends TestCase {
        private org.apache.xerces.dom.CoreDocumentImpl document = null;
        private XSchemaProcessor processor = null;
        private Vector rules = null;
        private org.apache.xerces.dom.ElementImpl testNode = null;
        
        public TEST(String testName) {
            super(testName);
        }
        
        public static Test suite() {
            return ( new TestSuite(TEST.class));
        }
        
        protected void setUp() throws Exception {
            super.setUp();
            processor = new XSchemaProcessor();
            rules = new Vector();
            Rule testRule = new Rule("test","test");
            testRule.addCondition(new siena.xml.rules.ExistsCondition(true));
            Rule testAttRule = new Rule("test@attr", "testattr");
            testAttRule.addCondition(new siena.xml.rules.ExistsCondition(true));
            Rule tempRule = new Rule("temp","temp");
            tempRule.addCondition(new siena.xml.rules.ExistsCondition(true));
            Rule childRule = new Rule("child","child");
            childRule.addCondition(new siena.xml.rules.ExistsCondition(true));
            Rule parentRule = new Rule("parent","parent");
            parentRule.addCondition(new siena.xml.rules.ExistsCondition(true));
            rules.addElement(testRule);
            rules.addElement(testAttRule);
            rules.addElement(tempRule);
            rules.addElement(childRule);
            rules.addElement(parentRule);
            document = new org.apache.xerces.dom.CoreDocumentImpl();
        }
        
        public void testUseRulesNoRulesApplyTest() {
            System.out.println(" - Test: XSchemaProcessor - useRules method, no rules apply to node");
            testNode = (ElementImpl)document.createElementNS("","norules");
            Assert.assertTrue("useRules returned true when no rules applied.",!processor.useRules(rules, testNode));
        }
        
        public void testUseRulesOneRuleAppliesTest() {
            System.out.println(" - Test: XSchemaProcessor - useRules method, one rule applied to node");
            testNode = (ElementImpl)document.createElementNS("","test");
            testNode.setAttribute("name","test");
            Assert.assertTrue("useRules returned false when a rule could be applied to the node.",processor.useRules(rules, testNode));
        }
        
        public void testUseRulesMultiRulesApplyTest() {
            System.out.println(" - Test: XSchemaProcessor - useRules method, multiple rules apply to node");
            testNode = (ElementImpl)document.createElementNS("","test");
            testNode.setAttribute("attr", "exists");
            Assert.assertTrue("useRules returned false when multiple rules could be applied to the node.", processor.useRules(rules, testNode));
        }
        
        public void testApplyRuleUniqueTest() {
            java.io.StringWriter writer = null;
            java.io.PrintWriter printWriter = null;
            
            Rule uniqueActionRule = new Rule("test", "unique:base");
            
            System.out.println(" - Test: XSchemaProcessor - applyRule method, unique processing insturction returned from rule.");
            
            testNode = (ElementImpl)document.createElementNS("","test");
            testNode.setAttribute("name","testNode");
            
            try {
                processor.applyRule(uniqueActionRule, testNode);
            } catch (Exception exception) {
                writer = new java.io.StringWriter();
                printWriter = new java.io.PrintWriter(writer);
                writer.write("Exception caught during test run: ");
                writer.write("\n\n*******BEGIN EXCEPTION********\n\n");
                exception.printStackTrace(printWriter);
                writer.write("\n*******END EXCEPTION********\n");
                fail(writer.toString());
            }
            
            Object result = processor.currentRules.get("base");
            Assert.assertNotNull("The result should not be null after the method call.", result);
            Assert.assertEquals("The value of the result is not correct.", "base/$testNode",result);
        }
        
        public void testApplyRuleUniqueCurrentTypeTest() {
            java.io.StringWriter writer = null;
            java.io.PrintWriter printWriter = null;
            Object result = null;
            Rule uniqueActionRule = new Rule("test", "unique:currentTag");
            
            System.out.println(" - Test: XSchemaProcessor - applyRule method, unique processing instruction specifying currentTag");
            
            testNode = (ElementImpl)document.createElementNS("","test");
            testNode.setAttribute("name","testNode");
            processor.currentType = "testType";
            
            try {
                processor.applyRule(uniqueActionRule, testNode);
            } catch (Exception exception) {
                writer = new java.io.StringWriter();
                printWriter = new java.io.PrintWriter(writer);
                writer.write("Exception caught during test run: ");
                writer.write("\n\n*******BEGIN EXCEPTION********\n\n");
                exception.printStackTrace(printWriter);
                writer.write("\n*******END EXCEPTION********\n");
                fail(writer.toString());
            }
            
            result = processor.currentRules.get(processor.currentType);
            Assert.assertNotNull("Result is null when it should not be.", result);
            Assert.assertEquals("Result value is not correct.", processor.currentType+"/$testNode", result);
        }
        
        public void testApplyRuleNoNameTest() {
            java.io.StringWriter writer = null;
            java.io.PrintWriter printWriter = null;
            
            Rule aRule = new Rule("test", "test");
            System.out.println(" - Test: XSchemaProcessor - applyRule method, attribute called name for the node");
            
            testNode = (ElementImpl)document.createElementNS("","test");
            
            processor.currentRules = new HashMap();
            
            try {
                processor.applyRule(aRule,testNode);
            } catch (siena.SienaException exception) {
                // Return since we were expecting an exception to be thrown here.
                return;
            }
            
            Assert.assertTrue("A rule was created when there should not have been a rule.", processor.currentRules.isEmpty());
        }
        
        public void testApplyRuleNameValueTest() {
            java.io.StringWriter writer = null;
            java.io.PrintWriter printWriter = null;
            
            Rule aRule = new Rule("test", "test");
            
            System.out.println(" - Test: XSchemaProcessor - applyRule method");
            
            testNode = (ElementImpl)document.createElementNS("","test");
            testNode.setAttribute("name","testName");
            
            processor.currentRules = new HashMap();
            
            try {
                processor.applyRule(aRule,testNode);
            } catch (Exception exception) {
                writer = new java.io.StringWriter();
                printWriter = new java.io.PrintWriter(writer);
                writer.write("Exception caught during test run: ");
                writer.write("\n\n*******BEGIN EXCEPTION********\n\n");
                exception.printStackTrace(printWriter);
                writer.write("\n*******END EXCEPTION********\n");
                fail(writer.toString());
            }
            
            Object result = processor.currentRules.get("testName");
            Assert.assertNotNull("Result was null when it should not have been.", result);
        }
        
        public void testGenerateXMLRulesElementTypeTest() {
            java.io.StringWriter writer = null;
            java.io.PrintWriter printWriter = null;
            
            System.out.println(" - Test: XSchemaProcessor - generateXMLRules method, currentType is set to element type's name");
            
            Node childNode = (ElementImpl)document.createElementNS("", "child");
            testNode = (ElementImpl)document.createElementNS("","element");
            testNode.setAttribute("name","passed");
            
            try {
                testNode.appendChild(childNode);
            } catch (Exception exception) {
                writer = new java.io.StringWriter();
                printWriter = new java.io.PrintWriter(writer);
                writer.write("Exception caught during test run: ");
                writer.write("\n\n*******BEGIN EXCEPTION********\n\n");
                exception.printStackTrace(printWriter);
                writer.write("\n*******END EXCEPTION********\n");
                fail(writer.toString());
            }
            
            processor.generateXMLRules(testNode, "failed");
            
            Assert.assertEquals("generateXMLRules did not set the correct type.", "passed", processor.currentType);
        }
        
        public void testGenerateXMLRulesComplexTypeTest() {
            java.io.StringWriter writer = null;
            java.io.PrintWriter printWriter = null;
            
            System.out.println(" - Test: XSchemaProcessor - generateXMLRules method, currentType is set to complex type name");
            
            Node childNode = (ElementImpl)document.createElementNS("", "child");
            testNode = (ElementImpl)document.createElementNS("","complexType");
            testNode.setAttribute("name","passed");
            
            try {
                testNode.appendChild(childNode);
            } catch (Exception exception) {
                writer = new java.io.StringWriter();
                printWriter = new java.io.PrintWriter(writer);
                writer.write("Exception caught during test run: ");
                writer.write("\n\n*******BEGIN EXCEPTION********\n\n");
                exception.printStackTrace(printWriter);
                writer.write("\n*******END EXCEPTION********\n");
                fail(writer.toString());
            }
            
            processor.generateXMLRules(testNode, "failed");
            
            Assert.assertEquals("generateXMLRuels did not get the correct complex type.", "passed", processor.currentType);
        }
        
        public void testGenerateXMLRulesRuleExistTest() {
            System.out.println(" - Test: XSchemaProcessor - generateXMLRules method, a rule exists for the node");
            processor.defaultRules.put("test", rules);
            
            testNode = (ElementImpl)document.createElementNS("","test");
            testNode.setAttribute("name","passed");
            
            processor.generateXMLRules(testNode, "");
            
            String result = (String)processor.currentRules.get("passed");
            
            Assert.assertEquals("The result dose not have the right value.", "test", result);
        }
        
        public void testGenerateXMLRulesAnyRuleTest() {
            System.out.println(" - Test: XSchemaProcessor - generateXMLRules method, use of the \"any\" rules");
            processor.defaultRules.put("any", rules);
            
            testNode = (ElementImpl)document.createElementNS("","test");
            testNode.setAttribute("name","passed");
            
            processor.generateXMLRules(testNode, "");
            
            String result = (String)processor.currentRules.get("passed");
            
            Assert.assertEquals("generateXMLRules did not apply the \"any\" rule to the node.", "test", result);
        }
        
        public void testGetRule() {
            System.out.println(" - Test: XSchemaProcessor - getRule method");
            testNode = (ElementImpl)document.createElementNS("","norule");
            
            Assert.assertNull("The getRule method returned a rule for a null parameter.", processor.getRule(null, processor.defaultRules));
            Assert.assertNull("The getRule method returned a rule when it should not have.", processor.getRule(testNode, processor.defaultRules));
        }
        
        public static void main(String[] args) {
            junit.textui.TestRunner.run(suite());
        }
    }
}
