//
//  This file is part of SienaXML, an XML interface to the 
//  Siena wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Author: Chris Corliss <corliss@falcon.stortek.com>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2002 University of Colorado
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//  
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//  
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: EqualityCondition.java,v 1.1.1.1 2002/06/07 22:42:37 jgiacomo Exp $
//
/* Generated by Together */

package siena.xml.rules;

import org.w3c.dom.Node;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Attr;
import java.util.StringTokenizer;

/** This condition represents one that checks for an equality.
 */
public class EqualityCondition implements Condition {
    /** Flag that indicates if equality or non-equality should be tested for.
     */    
    private boolean isEquals = true;
    /** Condition text to check objects against
     */    
    private String condition;
    /** The tree axis which to look at when doing the comparison
     */    
    private String treeAxis = null;

    /**
     * public constructor.
     *
     * @param shouldBeEqual Sets if the condition should check for equality or non-equality.
     * @param condition The string that contains the condition which to test for
     */
    public EqualityCondition(String condition, boolean shouldBeEqual) {
        initialize(condition, shouldBeEqual);
    }

    /** Constructor
     * @param value Value to check equality for
     * @param shouldBeEqual Sets if the condition should check for equality or non-equality.
     */    
    public EqualityCondition(int value, boolean shouldBeEqual) {
        initialize(value+"", shouldBeEqual); 
    }
    
    /** Constructor
     * @param value Value to check equality for
     * @param shouldBeEqual Sets if the condition should check for equality or non-equality.
     */    
    public EqualityCondition(double value, boolean shouldBeEqual) {
        initialize(value+"", shouldBeEqual);
    }
    
    /** This will initialize the object so that it can handle the comparison of any object correctly.
     * @param condition The string condition to check for
     * @param shouldBeEqual If equality or non-equality should be checked for.
     */    
    private void initialize(String condition, boolean shouldBeEqual) {
        StringTokenizer tokens = new StringTokenizer(condition);
        if (tokens.countTokens() > 1) {
            treeAxis = tokens.nextToken();
        }
        this.condition = tokens.nextToken();
        isEquals = shouldBeEqual;
    }
    /**
     * This method determins if the object passed in is equal to the condition
     * represented by this object.
     *
     * @param whatToEvaluate Object to test to see if it meets the condition
     * @return True if the object is equal to the condition String, otherwise
     * false.
     */
    
    public boolean meetCondition(Object whatToEvaluate) {
        
        String value = null;

        if (whatToEvaluate == null) {
            return false;
        }

        if (whatToEvaluate instanceof Node) {
            value = siena.xml.DOMTreeProcessing.getNodeValue((Node)whatToEvaluate);
            if (treeAxis != null) {
                return checkTreeCondition((Node)whatToEvaluate);
            }
        } else {
            value = whatToEvaluate.toString();
        }
        
        if (value == null) {
            return false;
        }
        
        if (isEquals && value.equalsIgnoreCase(condition)) {
            return true;
        } else if(!isEquals && !value.equalsIgnoreCase(condition)) {
            return true;
        }
        return false;
    }

    /** This method checks to see if a tree condition is satisfied by the Node
     * @param node Node used as the starting point for the tree axis.
     * @return True if the tree axis condition is met by this node.
     */    
    private boolean checkTreeCondition(Node node) {
        Node startingNode = null;
        Node temp = null;
        Node testNode = null;
        int counter = 0;
        boolean workWithAttributes = false;
        boolean result = false;
        StringTokenizer tokens = null;
        String targetTag = null;
        String targetAttr = null;
        String targetValue = null;
        String nodeValue = null;
        
        tokens = new StringTokenizer(condition, "@");
        targetTag = tokens.nextToken();
        targetAttr = null;
        targetValue = siena.xml.DOMTreeProcessing.getNodeValue(node);

        if (node instanceof Attr) {
            startingNode = ((Attr)node).getOwnerElement();
        } else {
            startingNode = node;
        }
        
        if (tokens.hasMoreTokens()) {
            targetAttr = tokens.nextToken();
            targetValue = startingNode.getAttributes().getNamedItem(targetAttr).getNodeValue();
            workWithAttributes = true;
        }
        
        if (treeAxis.equals("sibling")) {
            
            /**
             * First check to see if the target Tag is the keyword \"any\". If it is not
             * then check to see if there are any other siblings of the node passed in
             * that have the specified tag. If not, i.e. the counter is zero, the return
             * false since the tag portion of the condition did not work.
             */
            
            if (!targetTag.equals("any")) {
                counter = siena.xml.DOMTreeProcessing.countSiblingsWithTag(startingNode,targetTag);
                if (counter == 0 && isEquals) {
                    return (false | !isEquals);
                }
            }
            
            /**
             * If an attribute was specified in part of the condition then check to see
             * if any of the siblings have the same attribute with the same value. If the
             * condition is specified as an equality, then return true if the counter is
             * greater than zero. If the condition was specified as not equals and the counter is 
             * equal to zero then return zero otherwise return false.
             */
            
            if (workWithAttributes) {
                counter = siena.xml.DOMTreeProcessing.countSameAttributeValueSiblings(startingNode, targetAttr);
                if (counter > 0 && isEquals) {
                    return true;
                } else if (counter == 0 && !isEquals) {
                    return true;
                } else {
                    return false;
                }
            }

            temp = startingNode.getPreviousSibling();
            while (temp != null) {
                nodeValue = siena.xml.DOMTreeProcessing.getNodeValue(temp);
                if (nodeValue != null && nodeValue.equals(targetValue)) {
                    result = true;
                }
                temp = temp.getPreviousSibling();
            }
            temp = startingNode.getNextSibling();
            while (temp != null) {
                nodeValue = siena.xml.DOMTreeProcessing.getNodeValue(temp);
                if (nodeValue != null && nodeValue.equals(targetValue)) {
                    result = true;
                }
                temp = temp.getNextSibling();
            }
        } else if (treeAxis.equals("parent")) {
            temp = startingNode.getParentNode();
            
            // If the condition did not specify the keyword "any" then check to see if the 
            // parent tag matches the target tag specified in the condition.
            if (!targetTag.equals("any") && !targetTag.equals(temp.getLocalName())) {
                result = false;
            }
            
            if (workWithAttributes) {
                testNode = getNamedAttribute(targetAttr, temp);
                if (testNode != null && testNode.getNodeValue().equals(targetValue)) {
                    result = true;
                }
            } else {
                nodeValue = siena.xml.DOMTreeProcessing.getNodeValue(temp);                
                if (nodeValue != null && nodeValue.equals(targetValue)) {
                    result = true;
                }
            }
        } 
        
        /**
         * If the condition is checking for equality then simply return the result.
         * If the condition is checking for inequality then return the reverse of the
         * result.
         */
        
        if (isEquals) {
            return result;
        } else {
            return !result;
        }
    }
        
    /** Gets the value of the attribute with name <I>name</I> from a node
     * @param attrName Name of the attribute to retrieve
     * @param node Node to extract the attribute from.
     * @return Value of the attribute or null if the attribute does not exist on this node.
     */    
    private Node getNamedAttribute (String attrName, Node node) {
        NamedNodeMap attributes = node.getAttributes();
        Node attributeNode = null;
        if (attributes != null) {
            attributeNode = attributes.getNamedItem(attrName);
            return attributeNode;
        }               
        return attributeNode;
    }
}
